@@ -20,7 +20,6 @@ details.  */
 #include <java/lang/Thread.h>
 #include <java/lang/ThreadGroup.h>
 #include <java/lang/IllegalArgumentException.h>
-#include <java/lang/UnsupportedOperationException.h>
 #include <java/lang/IllegalThreadStateException.h>
 #include <java/lang/InterruptedException.h>
 #include <java/lang/NullPointerException.h>
@@ -91,8 +90,10 @@ jint
 java::lang::Thread::countStackFrames (void)
 {
   // NOTE: This is deprecated in JDK 1.2.
-  throw new UnsupportedOperationException
-    (JvNewStringLatin1 ("Thread.countStackFrames unimplemented"));
+
+  // Old applets still call this method.  Rather than throwing
+  // UnsupportedOperationException we simply fail silently.
+
   return 0;
 }
 
@@ -150,8 +151,9 @@ void
 java::lang::Thread::resume (void)
 {
   checkAccess ();
-  throw new UnsupportedOperationException
-    (JvNewStringLatin1 ("Thread.resume unimplemented"));
+
+  // Old applets still call this method.  Rather than throwing
+  // UnsupportedOperationException we simply fail silently.
 }
 
 void
@@ -216,6 +218,9 @@ java::lang::Thread::finish_ ()
   // If a method cache was created, free it.
   _Jv_FreeMethodCache();
 
+  // Clear out thread locals.
+  locals = NULL;
+
   // Signal any threads that are waiting to join() us.
   _Jv_MutexLock (&nt->join_mutex);
 
@@ -313,6 +318,13 @@ _Jv_ThreadRun (java::lang::Thread* thread)
   thread->finish_ ();
 }
 
+_Jv_Thread_t*
+_Jv_ThreadGetData (java::lang::Thread* thread)
+{
+  natThread* nt = (natThread*) thread->data;
+  return nt->thread;
+}
+
 void
 java::lang::Thread::start (void)
 {
@@ -332,16 +344,18 @@ void
 java::lang::Thread::stop (java::lang::Throwable *)
 {
   checkAccess ();
-  throw new UnsupportedOperationException
-    (JvNewStringLatin1 ("Thread.stop unimplemented"));
+
+  // Old applets still call this method.  Rather than throwing
+  // UnsupportedOperationException we simply fail silently.
 }
 
 void
 java::lang::Thread::suspend (void)
 {
   checkAccess ();
-  throw new UnsupportedOperationException 
-    (JvNewStringLatin1 ("Thread.suspend unimplemented"));
+
+  // Old applets still call this method.  Rather than throwing
+  // UnsupportedOperationException we simply fail silently.
 }
 
 static int nextThreadNumber = 0;
@@ -396,7 +410,8 @@ _Jv_SetCurrentJNIEnv (JNIEnv *env)
 }
 
 // Attach the current native thread to an existing (but unstarted) Thread 
-// object. Returns -1 on failure, 0 upon success.
+// object. Does not register thread with the garbage collector.
+// Returns -1 on failure, 0 upon success.
 jint
 _Jv_AttachCurrentThread(java::lang::Thread* thread)
 {
@@ -413,6 +428,8 @@ _Jv_AttachCurrentThread(java::lang::Thread* thread)
 java::lang::Thread*
 _Jv_AttachCurrentThread(jstring name, java::lang::ThreadGroup* group)
 {
+  // Register thread with GC before attempting any allocations.
+  _Jv_GCAttachThread ();
   java::lang::Thread *thread = _Jv_ThreadCurrent ();
   if (thread != NULL)
     return thread;
@@ -447,6 +464,7 @@ _Jv_DetachCurrentThread (void)
     return -1;
 
   _Jv_ThreadUnRegister ();
+  _Jv_GCDetachThread ();
   // Release the monitors.
   t->finish_ ();
 
